package userpasswd

import (
	"context"
	"encoding/json"
	"net/http"
	"net/url"

	"github.com/cockroachdb/errors"
	"github.com/inconshreveable/log15"

	"github.com/sourcegraph/sourcegraph/cmd/frontend/backend"
	"github.com/sourcegraph/sourcegraph/cmd/frontend/globals"
	"github.com/sourcegraph/sourcegraph/internal/actor"
	"github.com/sourcegraph/sourcegraph/internal/conf"
	"github.com/sourcegraph/sourcegraph/internal/database"
	"github.com/sourcegraph/sourcegraph/internal/database/dbutil"
	"github.com/sourcegraph/sourcegraph/internal/errcode"
	"github.com/sourcegraph/sourcegraph/internal/txemail"
	"github.com/sourcegraph/sourcegraph/internal/txemail/txtypes"
)

func SendResetPasswordURLEmail(ctx context.Context, email, username string, resetURL *url.URL) error {
	return txemail.Send(ctx, txemail.Message{
		To:       []string{email},
		Template: resetPasswordEmailTemplates,
		Data: struct {
			Username string
			URL      string
			Host     string
		}{
			Username: username,
			URL:      globals.ExternalURL().ResolveReference(resetURL).String(),
			Host:     globals.ExternalURL().Host,
		},
	})
}

// HandleResetPasswordInit initiates the builtin-auth password reset flow by sending a password-reset email.
func HandleResetPasswordInit(db dbutil.DB) func(w http.ResponseWriter, r *http.Request) {
	return func(w http.ResponseWriter, r *http.Request) {
		if handleEnabledCheck(w) {
			return
		}
		if handleNotAuthenticatedCheck(w, r) {
			return
		}
		if !conf.CanSendEmail() {
			httpLogAndError(w, "Unable to reset password because email sending is not configured on this site", http.StatusNotFound)
			return
		}

		ctx := r.Context()
		var formData struct {
			Email string `json:"email"`
		}
		if err := json.NewDecoder(r.Body).Decode(&formData); err != nil {
			httpLogAndError(w, "Could not decode password reset request body", http.StatusBadRequest, "err", err)
			return
		}

		if formData.Email == "" {
			httpLogAndError(w, "No email specified in password reset request", http.StatusBadRequest)
			return
		}

		usr, err := database.Users(db).GetByVerifiedEmail(ctx, formData.Email)
		if err != nil {
			// 🚨 SECURITY: We don't show an error message when the user is not found
			// as to not leak the existence of a given e-mail address in the database.
			if !errcode.IsNotFound(err) {
				httpLogAndError(w, "Failed to lookup user", http.StatusInternalServerError)
			}
			return
		}

		resetURL, err := backend.MakePasswordResetURL(ctx, db, usr.ID)
		if err == database.ErrPasswordResetRateLimit {
			httpLogAndError(w, "Too many password reset requests. Try again in a few minutes.", http.StatusTooManyRequests, "err", err)
			return
		} else if err != nil {
			httpLogAndError(w, "Could not reset password", http.StatusBadRequest, "err", err)
			return
		}

		if err := SendResetPasswordURLEmail(r.Context(), formData.Email, usr.Username, resetURL); err != nil {
			httpLogAndError(w, "Could not send reset password email", http.StatusInternalServerError, "err", err)
			return
		}
		database.LogPasswordEvent(ctx, db, r, database.SecurityEventNamPasswordResetRequested, usr.ID)
	}
}

var resetPasswordEmailTemplates = txemail.MustValidate(txtypes.Templates{
	Subject: `Reset your Sourcegraph password ({{.Host}})`,
	Text: `
Somebody (likely you) requested a password reset for the user {{.Username}} on Sourcegraph ({{.Host}}).

To reset the password for {{.Username}} on Sourcegraph, follow this link:

  {{.URL}}
`,
	HTML: `
<p>
  Somebody (likely you) requested a password reset for <strong>{{.Username}}</strong>
  on Sourcegraph ({{.Host}}).
</p>

<p><strong><a href="{{.URL}}">Reset password for {{.Username}}</a></strong></p>
`,
})

// HandleSetPasswordEmail sends the password reset email directly to the user for users created by site admins.
func HandleSetPasswordEmail(ctx context.Context, db dbutil.DB, id int32) (string, error) {
	e, _, err := database.UserEmails(db).GetPrimaryEmail(ctx, id)
	if err != nil {
		return "", errors.Wrap(err, "get user primary email")
	}

	usr, err := database.Users(db).GetByID(ctx, id)
	if err != nil {
		return "", errors.Wrap(err, "get user by ID")
	}

	ru, err := backend.MakePasswordResetURL(ctx, db, id)
	if err == database.ErrPasswordResetRateLimit {
		return "", err
	} else if err != nil {
		return "", errors.Wrap(err, "make password reset URL")
	}

	rus := globals.ExternalURL().ResolveReference(ru).String()
	if err := txemail.Send(ctx, txemail.Message{
		To:       []string{e},
		Template: setPasswordEmailTemplates,
		Data: struct {
			Username string
			URL      string
			Host     string
		}{
			Username: usr.Username,
			URL:      rus,
			Host:     globals.ExternalURL().Host,
		},
	}); err != nil {
		return "", err
	}

	return rus, nil
}

var setPasswordEmailTemplates = txemail.MustValidate(txtypes.Templates{
	Subject: `Set your Sourcegraph password ({{.Host}})`,
	Text: `
Your administrator created an account for you on Sourcegraph ({{.Host}}).

To set the password for {{.Username}} on Sourcegraph, follow this link:

  {{.URL}}
`,
	HTML: `
<p>
  Your administrator created an account for you on Sourcegraph ({{.Host}}).
</p>

<p><strong><a href="{{.URL}}">Set password for {{.Username}}</a></strong></p>
`,
})

// HandleResetPasswordCode resets the password if the correct code is provided.
func HandleResetPasswordCode(db database.DB) func(w http.ResponseWriter, r *http.Request) {
	return func(w http.ResponseWriter, r *http.Request) {
		if handleEnabledCheck(w) {
			return
		}
		if handleNotAuthenticatedCheck(w, r) {
			return
		}

		ctx := r.Context()
		var params struct {
			UserID   int32  `json:"userID"`
			Code     string `json:"code"`
			Password string `json:"password"` // new password
		}
		if err := json.NewDecoder(r.Body).Decode(&params); err != nil {
			httpLogAndError(w, "Password reset with code: could not decode request body", http.StatusBadGateway, "err", err)
			return
		}

		if err := database.CheckPasswordLength(params.Password); err != nil {
			http.Error(w, err.Error(), http.StatusBadRequest)
			return
		}

		success, err := db.Users().SetPassword(ctx, params.UserID, params.Code, params.Password)
		if err != nil {
			httpLogAndError(w, "Unexpected error", http.StatusInternalServerError, "err", err)
			return
		}

		if !success {
			http.Error(w, "Password reset code was invalid or expired.", http.StatusUnauthorized)
			return
		}

		database.LogPasswordEvent(ctx, db, r, database.SecurityEventNamePasswordChanged, params.UserID)

		if conf.CanSendEmail() {
			if err := backend.UserEmails.SendUserEmailOnFieldUpdate(ctx, db, params.UserID, "reset the password"); err != nil {
				log15.Warn("Failed to send email to inform user of password reset", "error", err)
			}
		}
	}
}

func handleNotAuthenticatedCheck(w http.ResponseWriter, r *http.Request) (handled bool) {
	if actor.FromContext(r.Context()).IsAuthenticated() {
		http.Error(w, "Authenticated users may not perform password reset.", http.StatusInternalServerError)
		return true
	}
	return false
}
