package main

import (
	"context"
	"fmt"

	"github.com/cockroachdb/errors"
	"github.com/google/go-cmp/cmp"
)

// buildQueries returns a channel that is fed all of the test functions that should be invoked
// as part of the test. This function depends on the flags provided by the user to alter the
// behavior of the testing functions.
func buildQueries() <-chan queryFunc {
	fns := make(chan queryFunc)

	go func() {
		defer close(fns)

		for _, testCase := range testCases {
			// Definition returns defintion
			fns <- makeTestFunc("def -> def", queryDefinitions, testCase.Definition, []Location{testCase.Definition})

			// References return definition
			for _, reference := range testCase.References {
				fns <- makeTestFunc("refs -> def", queryDefinitions, reference, []Location{testCase.Definition})
			}

			// Definition returns references
			fns <- makeTestFunc("def -> refs", queryReferences, testCase.Definition, testCase.References)

			// References return references
			if queryReferencesOfReferences {
				for _, reference := range testCase.References {
					fns <- makeTestFunc("refs -> refs", queryReferences, reference, testCase.References)
				}
			}
		}
	}()

	return fns
}

type testFunc func(ctx context.Context, location Location) ([]Location, error)

// makeTestFunc returns a test function that invokes the given function f with the given
// source, then compares the result against the set of expected locations. This function
// depends on the flags provided by the user to alter the behavior of the testing
// functions.
func makeTestFunc(name string, f testFunc, source Location, expectedLocations []Location) func(ctx context.Context) error {
	return func(ctx context.Context) error {
		locations, err := f(ctx, source)
		if err != nil {
			return err
		}

		if checkQueryResult {
			sortLocations(locations)

			if diff := cmp.Diff(expectedLocations, locations); diff != "" {
				collectRepositoryToResults := func(locations []Location) map[string]int {
					repositoryToResults := map[string]int{}
					for _, location := range locations {
						if _, ok := repositoryToResults[location.Repo]; !ok {
							repositoryToResults[location.Repo] = 0
						}
						repositoryToResults[location.Repo] += 1
					}
					return repositoryToResults
				}

				e := ""
				e += fmt.Sprintf("%s: unexpected results\n\n", name)
				e += fmt.Sprintf("started at location:\n\n    %+v\n\n", source)
				e += "results by repository:\n\n"

				allRepos := map[string]struct{}{}
				for _, location := range append(locations, expectedLocations...) {
					allRepos[location.Repo] = struct{}{}
				}
				repositoryToGottenResults := collectRepositoryToResults(locations)
				repositoryToWantedResults := collectRepositoryToResults(expectedLocations)
				for repo := range allRepos {
					e += fmt.Sprintf("    - %s: want %d got %d locations\n", repo, repositoryToWantedResults[repo], repositoryToGottenResults[repo])
				}
				e += "\n"

				e += "raw diff (-want +got):\n\n" + diff

				return errors.Errorf(e)
			}
		}

		return nil
	}
}
