package database

import (
	"context"
	"database/sql"
	"testing"
	"time"

	"github.com/keegancsmith/sqlf"
	"github.com/stretchr/testify/require"

	"github.com/sourcegraph/sourcegraph/internal/actor"
	"github.com/sourcegraph/sourcegraph/internal/database/basestore"
	"github.com/sourcegraph/sourcegraph/internal/database/dbtest"
	ff "github.com/sourcegraph/sourcegraph/internal/featureflag"
	"github.com/sourcegraph/sourcegraph/internal/types"
)

func TestFeatureFlagStore(t *testing.T) {
	t.Parallel()
	t.Run("NewFeatureFlag", testNewFeatureFlagRoundtrip)
	t.Run("ListFeatureFlags", testListFeatureFlags)
	t.Run("Overrides", func(t *testing.T) {
		t.Run("NewOverride", testNewOverrideRoundtrip)
		t.Run("ListUserOverrides", testListUserOverrides)
		t.Run("ListOrgOverrides", testListOrgOverrides)
	})
	t.Run("UserFlags", testUserFlags)
	t.Run("AnonymousUserFlags", testAnonymousUserFlags)
	t.Run("UserlessFeatureFlags", testUserlessFeatureFlags)
	t.Run("OrganizationFeatureFlag", testOrgFeatureFlag)
}

func errorContains(s string) require.ErrorAssertionFunc {
	return func(t require.TestingT, err error, msg ...interface{}) {
		require.Error(t, err)
		require.Contains(t, err.Error(), s, msg)
	}
}

func cleanup(t *testing.T, db *sql.DB) func() {
	return func() {
		if t.Failed() {
			// Retain content on failed tests
			return
		}
		_, err := db.Exec(`truncate feature_flags, feature_flag_overrides, users, orgs, org_members cascade;`)
		require.NoError(t, err)
	}
}

func testNewFeatureFlagRoundtrip(t *testing.T) {
	t.Parallel()
	flagStore := FeatureFlags(dbtest.NewDB(t))
	ctx := actor.WithInternalActor(context.Background())

	cases := []struct {
		flag      *ff.FeatureFlag
		assertErr require.ErrorAssertionFunc
	}{
		{
			flag: &ff.FeatureFlag{Name: "bool_true", Bool: &ff.FeatureFlagBool{Value: true}},
		},
		{
			flag: &ff.FeatureFlag{Name: "bool_false", Bool: &ff.FeatureFlagBool{Value: false}},
		},
		{
			flag: &ff.FeatureFlag{Name: "min_rollout", Rollout: &ff.FeatureFlagRollout{Rollout: 0}},
		},
		{
			flag: &ff.FeatureFlag{Name: "mid_rollout", Rollout: &ff.FeatureFlagRollout{Rollout: 3124}},
		},
		{
			flag: &ff.FeatureFlag{Name: "max_rollout", Rollout: &ff.FeatureFlagRollout{Rollout: 10000}},
		},
		{
			flag:      &ff.FeatureFlag{Name: "err_too_high_rollout", Rollout: &ff.FeatureFlagRollout{Rollout: 10001}},
			assertErr: errorContains(`violates check constraint "feature_flags_rollout_check"`),
		},
		{
			flag:      &ff.FeatureFlag{Name: "err_too_low_rollout", Rollout: &ff.FeatureFlagRollout{Rollout: -1}},
			assertErr: errorContains(`violates check constraint "feature_flags_rollout_check"`),
		},
		{
			flag:      &ff.FeatureFlag{Name: "err_no_types"},
			assertErr: errorContains(`feature flag must have exactly one type`),
		},
	}

	for _, tc := range cases {
		t.Run(tc.flag.Name, func(t *testing.T) {
			res, err := flagStore.CreateFeatureFlag(ctx, tc.flag)
			if tc.assertErr != nil {
				tc.assertErr(t, err)
				return
			}
			require.NoError(t, err)

			// Only assert that the values it is created with are equal.
			// Don't bother with the timestamps
			require.Equal(t, tc.flag.Name, res.Name)
			require.Equal(t, tc.flag.Bool, res.Bool)
			require.Equal(t, tc.flag.Rollout, res.Rollout)
		})
	}
}

func testListFeatureFlags(t *testing.T) {
	t.Parallel()
	flagStore := &featureFlagStore{Store: basestore.NewWithDB(dbtest.NewDB(t), sql.TxOptions{})}
	ctx := actor.WithInternalActor(context.Background())

	flag1 := &ff.FeatureFlag{Name: "bool_true", Bool: &ff.FeatureFlagBool{Value: true}}
	flag2 := &ff.FeatureFlag{Name: "bool_false", Bool: &ff.FeatureFlagBool{Value: false}}
	flag3 := &ff.FeatureFlag{Name: "mid_rollout", Rollout: &ff.FeatureFlagRollout{Rollout: 3124}}
	flag4 := &ff.FeatureFlag{Name: "deletable", Rollout: &ff.FeatureFlagRollout{Rollout: 3125}}
	flags := []*ff.FeatureFlag{flag1, flag2, flag3, flag4}

	for _, flag := range flags {
		_, err := flagStore.CreateFeatureFlag(ctx, flag)
		require.NoError(t, err)
	}

	// Deleted flag4
	err := flagStore.Exec(ctx, sqlf.Sprintf("DELETE FROM feature_flags WHERE flag_name = 'deletable';"))
	require.NoError(t, err)

	expected := []*ff.FeatureFlag{flag1, flag2, flag3}

	res, err := flagStore.GetFeatureFlags(ctx)
	require.NoError(t, err)
	for _, flag := range res {
		// Unset any timestamps
		flag.CreatedAt = time.Time{}
		flag.UpdatedAt = time.Time{}
		flag.DeletedAt = nil
	}

	require.EqualValues(t, res, expected)
}

func testNewOverrideRoundtrip(t *testing.T) {
	t.Parallel()
	db := dbtest.NewDB(t)
	flagStore := FeatureFlags(db)
	users := Users(db)
	ctx := actor.WithInternalActor(context.Background())

	ff1, err := flagStore.CreateBool(ctx, "t", true)
	require.NoError(t, err)

	u1, err := users.Create(ctx, NewUser{Username: "u", Password: "p"})
	require.NoError(t, err)

	invalidUserID := int32(38535)

	cases := []struct {
		override  *ff.Override
		assertErr require.ErrorAssertionFunc
	}{
		{
			override: &ff.Override{UserID: &u1.ID, FlagName: ff1.Name, Value: false},
		},
		{
			override:  &ff.Override{UserID: &invalidUserID, FlagName: ff1.Name, Value: false},
			assertErr: errorContains(`violates foreign key constraint "feature_flag_overrides_namespace_user_id_fkey"`),
		},
		{
			override:  &ff.Override{UserID: &u1.ID, FlagName: "invalid-flag-name", Value: false},
			assertErr: errorContains(`violates foreign key constraint "feature_flag_overrides_flag_name_fkey"`),
		},
		{
			override:  &ff.Override{FlagName: ff1.Name, Value: false},
			assertErr: errorContains(`violates check constraint "feature_flag_overrides_has_org_or_user_id"`),
		},
	}

	for _, tc := range cases {
		t.Run("case", func(t *testing.T) {
			res, err := flagStore.CreateOverride(ctx, tc.override)
			if tc.assertErr != nil {
				tc.assertErr(t, err)
				return
			}
			require.NoError(t, err)
			require.Equal(t, tc.override, res)
		})
	}
}

func testListUserOverrides(t *testing.T) {
	t.Parallel()
	db := dbtest.NewDB(t)
	flagStore := &featureFlagStore{Store: basestore.NewWithDB(db, sql.TxOptions{})}
	users := Users(db)
	ctx := actor.WithInternalActor(context.Background())

	mkUser := func(name string) *types.User {
		u, err := users.Create(ctx, NewUser{Username: name, Password: "p"})
		require.NoError(t, err)
		return u
	}

	mkFFBool := func(name string, val bool) *ff.FeatureFlag {
		res, err := flagStore.CreateBool(ctx, name, val)
		require.NoError(t, err)
		return res
	}

	mkOverride := func(user int32, flag string, val bool) *ff.Override {
		ffo, err := flagStore.CreateOverride(ctx, &ff.Override{UserID: &user, FlagName: flag, Value: val})
		require.NoError(t, err)
		return ffo
	}

	t.Run("no overrides", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		u1 := mkUser("u")
		mkFFBool("f", true)
		got, err := flagStore.GetUserOverrides(ctx, u1.ID)
		require.NoError(t, err)
		require.Empty(t, got)
	})

	t.Run("some overrides", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		u1 := mkUser("u")
		f1 := mkFFBool("f", true)
		o1 := mkOverride(u1.ID, f1.Name, false)
		got, err := flagStore.GetUserOverrides(ctx, u1.ID)
		require.NoError(t, err)
		require.Equal(t, got, []*ff.Override{o1})
	})

	t.Run("overrides for other users", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		u1 := mkUser("u1")
		u2 := mkUser("u2")
		f1 := mkFFBool("f", true)
		o1 := mkOverride(u1.ID, f1.Name, false)
		mkOverride(u2.ID, f1.Name, true)
		got, err := flagStore.GetUserOverrides(ctx, u1.ID)
		require.NoError(t, err)
		require.Equal(t, got, []*ff.Override{o1})
	})

	t.Run("deleted override", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		u1 := mkUser("u1")
		f1 := mkFFBool("f", true)
		mkOverride(u1.ID, f1.Name, false)
		err := flagStore.Exec(ctx, sqlf.Sprintf("UPDATE feature_flag_overrides SET deleted_at = now()"))
		require.NoError(t, err)
		got, err := flagStore.GetUserOverrides(ctx, u1.ID)
		require.NoError(t, err)
		require.Empty(t, got)
	})

	t.Run("non-unique override errors", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		u1 := mkUser("u1")
		f1 := mkFFBool("f", true)
		_, err := flagStore.CreateOverride(ctx, &ff.Override{UserID: &u1.ID, FlagName: f1.Name, Value: true})
		require.NoError(t, err)
		_, err = flagStore.CreateOverride(ctx, &ff.Override{UserID: &u1.ID, FlagName: f1.Name, Value: true})
		require.Error(t, err)
	})
}

func testListOrgOverrides(t *testing.T) {
	t.Parallel()
	db := dbtest.NewDB(t)
	flagStore := &featureFlagStore{Store: basestore.NewWithDB(db, sql.TxOptions{})}
	users := Users(db)
	orgs := Orgs(db)
	orgMembers := OrgMembers(db)
	ctx := actor.WithInternalActor(context.Background())

	mkUser := func(name string, orgIDs ...int32) *types.User {
		u, err := users.Create(ctx, NewUser{Username: name, Password: "p"})
		require.NoError(t, err)
		for _, id := range orgIDs {
			_, err := orgMembers.Create(ctx, id, u.ID)
			require.NoError(t, err)
		}
		return u
	}

	mkFFBool := func(name string, val bool) *ff.FeatureFlag {
		res, err := flagStore.CreateBool(ctx, name, val)
		require.NoError(t, err)
		return res
	}

	mkOverride := func(org int32, flag string, val bool) *ff.Override {
		ffo, err := flagStore.CreateOverride(ctx, &ff.Override{OrgID: &org, FlagName: flag, Value: val})
		require.NoError(t, err)
		return ffo
	}

	mkOrg := func(name string) *types.Org {
		o, err := orgs.Create(ctx, name, nil)
		require.NoError(t, err)
		return o
	}

	t.Run("no overrides", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		u1 := mkUser("u")
		mkFFBool("f", true)

		got, err := flagStore.GetUserOverrides(ctx, u1.ID)
		require.NoError(t, err)
		require.Empty(t, got)
	})

	t.Run("some overrides", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		org1 := mkOrg("org1")
		u1 := mkUser("u", org1.ID)
		f1 := mkFFBool("f", true)
		o1 := mkOverride(org1.ID, f1.Name, false)

		got, err := flagStore.GetOrgOverridesForUser(ctx, u1.ID)
		require.NoError(t, err)
		require.Equal(t, got, []*ff.Override{o1})
	})

	t.Run("deleted overrides", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		org1 := mkOrg("org1")
		u1 := mkUser("u", org1.ID)
		f1 := mkFFBool("f", true)
		mkOverride(org1.ID, f1.Name, false)
		err := flagStore.Exec(ctx, sqlf.Sprintf("UPDATE feature_flag_overrides SET deleted_at = now();"))
		require.NoError(t, err)

		got, err := flagStore.GetOrgOverridesForUser(ctx, u1.ID)
		require.NoError(t, err)
		require.Empty(t, got)
	})

	t.Run("non-unique override errors", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		org1 := mkOrg("org1")
		f1 := mkFFBool("f", true)

		_, err := flagStore.CreateOverride(ctx, &ff.Override{OrgID: &org1.ID, FlagName: f1.Name, Value: true})
		require.NoError(t, err)
		_, err = flagStore.CreateOverride(ctx, &ff.Override{OrgID: &org1.ID, FlagName: f1.Name, Value: false})
		require.Error(t, err)
	})
}

func testUserFlags(t *testing.T) {
	t.Parallel()
	db := dbtest.NewDB(t)
	flagStore := FeatureFlags(db)
	users := Users(db)
	orgs := Orgs(db)
	orgMembers := OrgMembers(db)
	ctx := actor.WithInternalActor(context.Background())

	mkUser := func(name string, orgIDs ...int32) *types.User {
		u, err := users.Create(ctx, NewUser{Username: name, Password: "p"})
		require.NoError(t, err)
		for _, id := range orgIDs {
			_, err := orgMembers.Create(ctx, id, u.ID)
			require.NoError(t, err)
		}
		return u
	}

	mkFFBool := func(name string, val bool) *ff.FeatureFlag {
		res, err := flagStore.CreateBool(ctx, name, val)
		require.NoError(t, err)
		return res
	}

	mkFFBoolVar := func(name string, rollout int32) *ff.FeatureFlag {
		res, err := flagStore.CreateRollout(ctx, name, rollout)
		require.NoError(t, err)
		return res
	}

	mkUserOverride := func(user int32, flag string, val bool) *ff.Override {
		ffo, err := flagStore.CreateOverride(ctx, &ff.Override{UserID: &user, FlagName: flag, Value: val})
		require.NoError(t, err)
		return ffo
	}

	mkOrgOverride := func(org int32, flag string, val bool) *ff.Override {
		ffo, err := flagStore.CreateOverride(ctx, &ff.Override{OrgID: &org, FlagName: flag, Value: val})
		require.NoError(t, err)
		return ffo
	}

	mkOrg := func(name string) *types.Org {
		o, err := orgs.Create(ctx, name, nil)
		require.NoError(t, err)
		return o
	}

	t.Run("bool vals", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		u1 := mkUser("u")
		mkFFBool("f1", true)
		mkFFBool("f2", false)

		got, err := flagStore.GetUserFlags(ctx, u1.ID)
		require.NoError(t, err)
		expected := map[string]bool{"f1": true, "f2": false}
		require.Equal(t, expected, got)
	})

	t.Run("bool vars", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		u1 := mkUser("u")
		mkFFBoolVar("f1", 10000)
		mkFFBoolVar("f2", 0)

		got, err := flagStore.GetUserFlags(ctx, u1.ID)
		require.NoError(t, err)
		expected := map[string]bool{"f1": true, "f2": false}
		require.Equal(t, expected, got)
	})

	t.Run("bool vals with user override", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		u1 := mkUser("u")
		mkFFBool("f1", true)
		mkFFBool("f2", false)
		mkUserOverride(u1.ID, "f2", true)

		got, err := flagStore.GetUserFlags(ctx, u1.ID)
		require.NoError(t, err)
		expected := map[string]bool{"f1": true, "f2": true}
		require.Equal(t, expected, got)
	})

	t.Run("bool vars with user override", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		u1 := mkUser("u")
		mkFFBoolVar("f1", 10000)
		mkFFBoolVar("f2", 0)
		mkUserOverride(u1.ID, "f2", true)

		got, err := flagStore.GetUserFlags(ctx, u1.ID)
		require.NoError(t, err)
		expected := map[string]bool{"f1": true, "f2": true}
		require.Equal(t, expected, got)
	})

	t.Run("bool vals with org override", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		o1 := mkOrg("o1")
		u1 := mkUser("u", o1.ID)
		mkFFBool("f1", true)
		mkFFBool("f2", false)
		mkOrgOverride(o1.ID, "f2", true)

		got, err := flagStore.GetUserFlags(ctx, u1.ID)
		require.NoError(t, err)
		expected := map[string]bool{"f1": true, "f2": true}
		require.Equal(t, expected, got)
	})

	t.Run("bool vars with org override", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		o1 := mkOrg("o1")
		u1 := mkUser("u", o1.ID)
		mkFFBoolVar("f1", 10000)
		mkFFBoolVar("f2", 0)
		mkOrgOverride(o1.ID, "f2", true)

		got, err := flagStore.GetUserFlags(ctx, u1.ID)
		require.NoError(t, err)
		expected := map[string]bool{"f1": true, "f2": true}
		require.Equal(t, expected, got)
	})

	t.Run("user override beats org override", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		o1 := mkOrg("o1")
		u1 := mkUser("u", o1.ID)
		mkFFBoolVar("f1", 10000)
		mkFFBoolVar("f2", 0)
		mkOrgOverride(o1.ID, "f2", true)
		mkUserOverride(u1.ID, "f2", false)

		got, err := flagStore.GetUserFlags(ctx, u1.ID)
		require.NoError(t, err)
		expected := map[string]bool{"f1": true, "f2": false}
		require.Equal(t, expected, got)
	})
}

func testAnonymousUserFlags(t *testing.T) {
	t.Parallel()
	db := dbtest.NewDB(t)
	flagStore := FeatureFlags(db)
	ctx := actor.WithInternalActor(context.Background())

	mkFFBool := func(name string, val bool) *ff.FeatureFlag {
		res, err := flagStore.CreateBool(ctx, name, val)
		require.NoError(t, err)
		return res
	}

	mkFFBoolVar := func(name string, rollout int32) *ff.FeatureFlag {
		res, err := flagStore.CreateRollout(ctx, name, rollout)
		require.NoError(t, err)
		return res
	}

	t.Run("bool vals", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		mkFFBool("f1", true)
		mkFFBool("f2", false)

		got, err := flagStore.GetAnonymousUserFlags(ctx, "testuser")
		require.NoError(t, err)
		expected := map[string]bool{"f1": true, "f2": false}
		require.Equal(t, expected, got)
	})

	t.Run("bool vars", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		mkFFBoolVar("f1", 10000)
		mkFFBoolVar("f2", 0)

		got, err := flagStore.GetAnonymousUserFlags(ctx, "testuser")
		require.NoError(t, err)
		expected := map[string]bool{"f1": true, "f2": false}
		require.Equal(t, expected, got)
	})

	// No override tests for AnonymousUserFlags because no override
	// can be defined for an anonymous user.
}

func testUserlessFeatureFlags(t *testing.T) {
	t.Parallel()
	db := dbtest.NewDB(t)
	flagStore := FeatureFlags(db)
	ctx := actor.WithInternalActor(context.Background())

	mkFFBool := func(name string, val bool) *ff.FeatureFlag {
		res, err := flagStore.CreateBool(ctx, name, val)
		require.NoError(t, err)
		return res
	}

	mkFFBoolVar := func(name string, rollout int32) *ff.FeatureFlag {
		res, err := flagStore.CreateRollout(ctx, name, rollout)
		require.NoError(t, err)
		return res
	}

	t.Run("bool vals", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		mkFFBool("f1", true)
		mkFFBool("f2", false)

		got, err := flagStore.GetGlobalFeatureFlags(ctx)
		require.NoError(t, err)
		expected := map[string]bool{"f1": true, "f2": false}
		require.Equal(t, expected, got)
	})

	t.Run("bool vars", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		mkFFBoolVar("f1", 10000)
		mkFFBoolVar("f2", 0)

		got, err := flagStore.GetGlobalFeatureFlags(ctx)
		require.NoError(t, err)

		// Userless requests don't have a stable user to evaluate
		// bool variable flags, so none should be defined.
		//
		// TODO(camdencheek): consider evaluating rollout feature
		// flags with a static string so they are defined and stable,
		// but effectively statically random.
		expected := map[string]bool{}
		require.Equal(t, expected, got)
	})
}

func testOrgFeatureFlag(t *testing.T) {
	t.Parallel()
	db := dbtest.NewDB(t)
	flagStore := FeatureFlags(db)
	orgs := Orgs(db)
	ctx := actor.WithInternalActor(context.Background())

	mkFFBool := func(name string, val bool) *ff.FeatureFlag {
		res, err := flagStore.CreateBool(ctx, name, val)
		require.NoError(t, err)
		return res
	}

	mkOrgOverride := func(org int32, flag string, val bool) *ff.Override {
		ffo, err := flagStore.CreateOverride(ctx, &ff.Override{OrgID: &org, FlagName: flag, Value: val})
		require.NoError(t, err)
		return ffo
	}

	mkOrg := func(name string) *types.Org {
		o, err := orgs.Create(ctx, name, nil)
		require.NoError(t, err)
		return o
	}

	t.Run("bool vals", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		org := mkOrg("o")
		mkFFBool("f1", true)
		mkFFBool("f2", false)

		got1, err1 := flagStore.GetOrgFeatureFlag(ctx, org.ID, "f1")
		got2, err2 := flagStore.GetOrgFeatureFlag(ctx, org.ID, "f2")
		require.NoError(t, err1)
		require.NoError(t, err2)
		expected1 := true
		expected2 := false
		require.Equal(t, expected1, got1)
		require.Equal(t, expected2, got2)
	})

	t.Run("bool vals with org override", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		org1 := mkOrg("o1")
		org2 := mkOrg("o2")
		mkFFBool("f1", true)
		mkFFBool("f2", false)
		mkOrgOverride(org1.ID, "f1", false)
		mkOrgOverride(org1.ID, "f2", true)

		got, err := flagStore.GetOrgFeatureFlag(ctx, org1.ID, "f1")
		require.NoError(t, err)
		require.Equal(t, false, got)

		got, err = flagStore.GetOrgFeatureFlag(ctx, org1.ID, "f2")
		require.NoError(t, err)
		require.Equal(t, true, got)

		got, err = flagStore.GetOrgFeatureFlag(ctx, org2.ID, "f1")
		require.NoError(t, err)
		require.Equal(t, true, got)

		got, err = flagStore.GetOrgFeatureFlag(ctx, org2.ID, "f2")
		require.NoError(t, err)
		require.Equal(t, false, got)
	})

	t.Run("bool vals without flag defined", func(t *testing.T) {
		t.Cleanup(cleanup(t, db))
		org := mkOrg("o")

		got, err := flagStore.GetOrgFeatureFlag(ctx, org.ID, "f1")
		require.NoError(t, err)
		require.Equal(t, false, got)
	})
}
